"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.findPR = exports.getNewMajorVersion = exports.getNewMinorVersion = exports.getNewPatchVersion = exports.getCurrentVersion = exports.getDefaultBranch = exports.resolveConflicts = exports.closePR = exports.getChangedFilesForRebase = exports.getChangedFiles = exports.updatePr = exports.afterCreatePr = exports.isMergeable = exports.push = exports.commit = exports.abortMerge = exports.merge = exports.getRefDiff = exports.getDiff = exports.checkBranch = exports.clone = exports.branchConfig = exports.userConfig = exports.getApiHelper = void 0;
const core_1 = require("@actions/core");
const github_action_helper_1 = require("@technote-space/github-action-helper");
const misc_1 = require("./misc");
const variables_1 = require("./variables");
const { getWorkspace, getLocalRefspec, getRefspec } = github_action_helper_1.Utils;
const { getOctokit, ensureNotNullValue, useNpm } = github_action_helper_1.Utils;
const { getRepository, isPush } = github_action_helper_1.ContextHelper;
const getApiHelper = (octokit, context, logger) => new github_action_helper_1.ApiHelper(octokit, context.actionContext, logger);
exports.getApiHelper = getApiHelper;
const userConfig = (helper, context) => __awaiter(void 0, void 0, void 0, function* () {
    return yield helper.config(getWorkspace(), {
        name: (0, variables_1.getCommitName)(context),
        email: (0, variables_1.getCommitEmail)(context),
    });
});
exports.userConfig = userConfig;
const branchConfig = (helper, octokit, context) => __awaiter(void 0, void 0, void 0, function* () {
    return yield helper.config(getWorkspace(), {
        defaultBranch: yield (0, exports.getDefaultBranch)(octokit, context),
    });
});
exports.branchConfig = branchConfig;
const clone = (helper, logger, octokit, context) => __awaiter(void 0, void 0, void 0, function* () {
    const branchName = yield (0, variables_1.getPrBranchName)(helper, octokit, context);
    logger.startProcess('Fetching...');
    helper.useOrigin(true);
    yield helper.fetchOrigin(getWorkspace(), context.actionContext, ['--no-tags'], [getRefspec(branchName)]);
    yield helper.runCommand(getWorkspace(), {
        command: 'git reset',
        args: ['--hard'],
        suppressError: true,
    });
    logger.startProcess('Switching branch to [%s]...', branchName);
    yield helper.switchBranch(getWorkspace(), branchName);
});
exports.clone = clone;
const checkBranch = (helper, logger, octokit, context) => __awaiter(void 0, void 0, void 0, function* () {
    const clonedBranch = yield helper.getCurrentBranchName(getWorkspace());
    const branchName = yield (0, variables_1.getPrBranchName)(helper, octokit, context);
    if (branchName === clonedBranch) {
        yield helper.runCommand(getWorkspace(), {
            command: 'git merge',
            args: ['--no-edit', getLocalRefspec(branchName)],
        });
        yield helper.runCommand(getWorkspace(), 'ls -la');
        return !isPush(context.actionContext);
    }
    if (isPush(context.actionContext)) {
        throw new Error(`remote branch [${branchName}] not found.`);
    }
    logger.info('remote branch [%s] not found.', branchName);
    logger.info('now branch: %s', clonedBranch);
    const headRef = (0, misc_1.getPrHeadRef)(context);
    logger.startProcess('Cloning [%s] from the remote repo...', headRef);
    yield helper.fetchOrigin(getWorkspace(), context.actionContext, ['--no-tags'], [getRefspec(headRef)]);
    yield helper.switchBranch(getWorkspace(), headRef);
    yield helper.createBranch(getWorkspace(), branchName);
    yield helper.runCommand(getWorkspace(), 'ls -la');
    return false;
});
exports.checkBranch = checkBranch;
const getClearPackageCommands = (context) => {
    if ((0, misc_1.isDisabledDeletePackage)(context)) {
        return [];
    }
    return [
        'rm -f package.json',
        'rm -f package-lock.json',
        'rm -f yarn.lock',
    ];
};
const getGlobalInstallPackagesCommands = (context) => {
    const packages = (0, misc_1.getActionDetail)('globalInstallPackages', context, () => []);
    if (packages.length) {
        if (useNpm(getWorkspace(), (0, core_1.getInput)('PACKAGE_MANAGER'))) {
            return [
                'sudo npm install -g ' + packages.join(' '),
            ];
        }
        else {
            return [
                'sudo yarn global add ' + packages.join(' '),
            ];
        }
    }
    return [];
};
const getDevInstallPackagesCommands = (context) => {
    const packages = (0, misc_1.getActionDetail)('devInstallPackages', context, () => []);
    if (packages.length) {
        if (useNpm(getWorkspace(), (0, core_1.getInput)('PACKAGE_MANAGER'))) {
            return [
                'npm install --save-dev ' + packages.join(' '),
            ];
        }
        else {
            return [
                'yarn add --dev ' + packages.join(' '),
            ];
        }
    }
    return [];
};
const getInstallPackagesCommands = (context) => {
    const packages = (0, misc_1.getActionDetail)('installPackages', context, () => []);
    if (packages.length) {
        if (useNpm(getWorkspace(), (0, core_1.getInput)('PACKAGE_MANAGER'))) {
            return [
                'npm install --save ' + packages.join(' '),
            ];
        }
        else {
            return [
                'yarn add ' + packages.join(' '),
            ];
        }
    }
    return [];
};
const getExecuteCommands = (context) => (0, misc_1.getActionDetail)('executeCommands', context, () => []);
const getDiff = (helper, logger) => __awaiter(void 0, void 0, void 0, function* () {
    logger.startProcess('Checking diff...');
    yield helper.runCommand(getWorkspace(), 'git add --all');
    return yield helper.getDiff(getWorkspace());
});
exports.getDiff = getDiff;
const getRefDiff = (compare, helper, logger, context) => __awaiter(void 0, void 0, void 0, function* () {
    logger.startProcess('Checking references diff...');
    yield helper.fetchBranch(getWorkspace(), compare, context.actionContext);
    return (yield helper.getRefDiff(getWorkspace(), 'HEAD', compare, (0, misc_1.getGitFilterStatus)(context), '..'));
});
exports.getRefDiff = getRefDiff;
const initDirectory = (helper, logger) => __awaiter(void 0, void 0, void 0, function* () {
    logger.startProcess('Initializing working directory...');
    helper.useOrigin(true);
    yield helper.runCommand(getWorkspace(), { command: 'rm', args: ['-rdf', getWorkspace()] });
});
const merge = (branch, helper, logger, context) => __awaiter(void 0, void 0, void 0, function* () {
    logger.startProcess('Merging [%s] branch...', getLocalRefspec(branch));
    yield helper.fetchOrigin(getWorkspace(), context.actionContext, ['--no-tags'], [getRefspec(branch)]);
    yield (0, exports.userConfig)(helper, context);
    const results = yield helper.runCommand(getWorkspace(), {
        command: 'git merge',
        args: ['--no-edit', getLocalRefspec(branch)],
        suppressError: true,
    });
    return !results[0].stdout.some(RegExp.prototype.test, /^CONFLICT /);
});
exports.merge = merge;
const abortMerge = (helper, logger) => __awaiter(void 0, void 0, void 0, function* () {
    logger.startProcess('Aborting merge...');
    yield helper.runCommand(getWorkspace(), 'git merge --abort');
});
exports.abortMerge = abortMerge;
const commit = (helper, logger, context) => __awaiter(void 0, void 0, void 0, function* () {
    yield (0, exports.userConfig)(helper, context);
    logger.startProcess('Committing...');
    yield helper.makeCommit(getWorkspace(), (0, variables_1.getCommitMessage)(context));
});
exports.commit = commit;
const push = (branchName, helper, logger, context) => __awaiter(void 0, void 0, void 0, function* () {
    logger.startProcess('Pushing to %s@%s...', getRepository(context.actionContext), branchName);
    yield helper.push(getWorkspace(), branchName, context.actionContext);
});
exports.push = push;
const forcePush = (branchName, helper, logger, context) => __awaiter(void 0, void 0, void 0, function* () {
    logger.startProcess('Pushing to %s@%s...', getRepository(context.actionContext), branchName);
    yield helper.forcePush(getWorkspace(), branchName, context.actionContext);
});
const isMergeable = (number, octokit, context) => __awaiter(void 0, void 0, void 0, function* () {
    return (0, misc_1.getCache)((0, misc_1.getCacheKey)('pulls.get', {
        owner: context.actionContext.repo.owner,
        repo: context.actionContext.repo.repo,
        'pull_number': number,
    }), () => __awaiter(void 0, void 0, void 0, function* () {
        return ensureNotNullValue((yield octokit.rest.pulls.get({
            owner: context.actionContext.repo.owner,
            repo: context.actionContext.repo.repo,
            'pull_number': number,
        })).data.mergeable, false);
    }), context);
});
exports.isMergeable = isMergeable;
const afterCreatePr = (branchName, number, helper, logger, octokit, context) => __awaiter(void 0, void 0, void 0, function* () {
    var _a, _b, _c, _d;
    if ((_a = context.actionDetail.labels) === null || _a === void 0 ? void 0 : _a.length) {
        logger.info('Adding labels...');
        console.log(context.actionDetail.labels);
        yield octokit.rest.issues.addLabels(Object.assign(Object.assign({}, context.actionContext.repo), { 'issue_number': number, labels: context.actionDetail.labels }));
    }
    if ((_b = context.actionDetail.assignees) === null || _b === void 0 ? void 0 : _b.length) {
        logger.info('Adding assignees...');
        console.log(context.actionDetail.assignees);
        yield octokit.rest.issues.addAssignees(Object.assign(Object.assign({}, context.actionContext.repo), { 'issue_number': number, assignees: context.actionDetail.assignees }));
    }
    if (((_c = context.actionDetail.reviewers) === null || _c === void 0 ? void 0 : _c.length) || ((_d = context.actionDetail.teamReviewers) === null || _d === void 0 ? void 0 : _d.length)) {
        logger.info('Adding reviewers...');
        console.log(context.actionDetail.reviewers);
        console.log(context.actionDetail.teamReviewers);
        yield octokit.rest.pulls.requestReviewers(Object.assign(Object.assign({}, context.actionContext.repo), { 'pull_number': number, reviewers: context.actionDetail.reviewers, 'team_reviewers': context.actionDetail.teamReviewers }));
    }
    if ((0, misc_1.isActiveTriggerWorkflow)(context)) {
        // add empty commit to trigger pr event
        yield helper.runCommand(getWorkspace(), {
            command: 'git commit',
            args: [
                '--allow-empty',
                '-qm',
                (0, misc_1.getTriggerWorkflowMessage)(context),
            ],
        });
        yield (0, exports.push)(branchName, helper, logger, context);
    }
});
exports.afterCreatePr = afterCreatePr;
const updatePr = (branchName, files, output, helper, logger, octokit, context) => __awaiter(void 0, void 0, void 0, function* () {
    const apiHelper = (0, exports.getApiHelper)(getOctokit((0, misc_1.getApiToken)()), context, logger);
    const pr = yield apiHelper.findPullRequest(branchName);
    if (pr) {
        logger.startProcess('Creating comment to PullRequest...');
        yield apiHelper.createCommentToPr(branchName, yield (0, variables_1.getPrBody)(true, files, output, helper, octokit, context));
        return (0, exports.isMergeable)(pr.number, octokit, context);
    }
    logger.startProcess('Creating PullRequest...');
    const { number } = yield apiHelper.pullsCreate(branchName, {
        title: yield (0, variables_1.getPrTitle)(helper, octokit, context),
        body: yield (0, variables_1.getPrBody)(false, files, output, helper, octokit, context),
    });
    yield (0, exports.afterCreatePr)(branchName, number, helper, logger, octokit, context);
    return true;
});
exports.updatePr = updatePr;
const runCommand = (command, helper, logger, context) => __awaiter(void 0, void 0, void 0, function* () {
    if ('string' === typeof command) {
        return (yield helper.runCommand(getWorkspace(), command))[0];
    }
    const result = yield command(context, helper, logger);
    logger.displayCommand(result.command);
    if (result.stdout.length) {
        logger.displayStdout(result.stdout);
    }
    if (result.stderr.length) {
        logger.displayStderr(result.stderr);
    }
    return result;
});
const runCommands = (helper, logger, context) => __awaiter(void 0, void 0, void 0, function* () {
    const commands = [].concat.apply([], [
        getClearPackageCommands(context),
        getGlobalInstallPackagesCommands(context),
        getDevInstallPackagesCommands(context),
        getInstallPackagesCommands(context),
        getExecuteCommands(context),
    ]);
    logger.startProcess('Running commands...');
    const output = yield commands.reduce((prev, command) => __awaiter(void 0, void 0, void 0, function* () {
        const acc = yield prev;
        return acc.concat(yield runCommand(command, helper, logger, context));
    }), Promise.resolve([]));
    return {
        files: yield (0, exports.getDiff)(helper, logger),
        output,
    };
});
const getChangedFiles = (helper, logger, octokit, context) => __awaiter(void 0, void 0, void 0, function* () {
    yield (0, exports.clone)(helper, logger, octokit, context);
    if (yield (0, exports.checkBranch)(helper, logger, octokit, context)) {
        if (!(yield (0, exports.merge)((0, misc_1.getContextBranch)(context), helper, logger, context))) {
            yield (0, exports.abortMerge)(helper, logger);
            return { files: [], output: [], aborted: true };
        }
    }
    return runCommands(helper, logger, context);
});
exports.getChangedFiles = getChangedFiles;
const getChangedFilesForRebase = (helper, logger, octokit, context) => __awaiter(void 0, void 0, void 0, function* () {
    yield initDirectory(helper, logger);
    const branchName = (0, misc_1.getContextBranch)(context);
    yield helper.fetchOrigin(getWorkspace(), context.actionContext, ['--no-tags'], [getRefspec(branchName)]);
    yield helper.switchBranch(getWorkspace(), branchName);
    yield helper.createBranch(getWorkspace(), yield (0, variables_1.getPrBranchName)(helper, octokit, context));
    return runCommands(helper, logger, context);
});
exports.getChangedFilesForRebase = getChangedFilesForRebase;
const closePR = (branchName, logger, context, message) => __awaiter(void 0, void 0, void 0, function* () { return (0, exports.getApiHelper)(getOctokit((0, misc_1.getApiToken)()), context, logger).closePR(branchName, message !== null && message !== void 0 ? message : context.actionDetail.prCloseMessage); });
exports.closePR = closePR;
const resolveConflicts = (branchName, helper, logger, octokit, context) => __awaiter(void 0, void 0, void 0, function* () {
    if (yield (0, exports.merge)((0, misc_1.getContextBranch)(context), helper, logger, context)) {
        // succeeded to merge
        yield (0, exports.push)(branchName, helper, logger, context);
    }
    else {
        // failed to merge
        const { files, output } = yield (0, exports.getChangedFilesForRebase)(helper, logger, octokit, context);
        if (!files.length) {
            yield (0, exports.closePR)(branchName, logger, context);
            return 'has been closed because there is no diff';
        }
        yield (0, exports.commit)(helper, logger, context);
        yield forcePush(branchName, helper, logger, context);
        yield (0, exports.getApiHelper)(getOctokit((0, misc_1.getApiToken)()), context, logger).pullsCreateOrUpdate(branchName, {
            title: yield (0, variables_1.getPrTitle)(helper, octokit, context),
            body: yield (0, variables_1.getPrBody)(false, files, output, helper, octokit, context),
        });
    }
    return 'updated';
});
exports.resolveConflicts = resolveConflicts;
const getDefaultBranch = (octokit, context) => __awaiter(void 0, void 0, void 0, function* () {
    return (0, misc_1.getCache)((0, misc_1.getCacheKey)('repos', {
        owner: context.actionContext.repo.owner,
        repo: context.actionContext.repo.repo,
    }), () => (0, exports.getApiHelper)(octokit, context).getDefaultBranch(), context);
});
exports.getDefaultBranch = getDefaultBranch;
const getCurrentVersion = (octokit, context) => __awaiter(void 0, void 0, void 0, function* () { return (0, misc_1.getCache)((0, misc_1.getCacheKey)('current-version'), () => (0, exports.getApiHelper)(octokit, context).getLastTag(), context); });
exports.getCurrentVersion = getCurrentVersion;
const getNewPatchVersion = (octokit, context) => __awaiter(void 0, void 0, void 0, function* () { return (0, misc_1.getCache)((0, misc_1.getCacheKey)('new-patch-version'), () => (0, exports.getApiHelper)(octokit, context).getNewPatchVersion(), context); });
exports.getNewPatchVersion = getNewPatchVersion;
const getNewMinorVersion = (octokit, context) => __awaiter(void 0, void 0, void 0, function* () { return (0, misc_1.getCache)((0, misc_1.getCacheKey)('new-minor-version'), () => (0, exports.getApiHelper)(octokit, context).getNewMinorVersion(), context); });
exports.getNewMinorVersion = getNewMinorVersion;
const getNewMajorVersion = (octokit, context) => __awaiter(void 0, void 0, void 0, function* () { return (0, misc_1.getCache)((0, misc_1.getCacheKey)('new-major-version'), () => (0, exports.getApiHelper)(octokit, context).getNewMajorVersion(), context); });
exports.getNewMajorVersion = getNewMajorVersion;
const findPR = (branchName, octokit, context) => __awaiter(void 0, void 0, void 0, function* () { return (0, misc_1.getCache)((0, misc_1.getCacheKey)('pr', { branchName }), () => (0, exports.getApiHelper)(octokit, context).findPullRequest(branchName), context); });
exports.findPR = findPR;
